package calculus;
import java.util.Stack ;
import java.util.EmptyStackException ;

public class MathObject {
	private double n ;
	private boolean isFunc = false ;
	private boolean isX = false ;
	private boolean isOp = false ;
	private boolean isNum = false ;
	private boolean isUnary = false ;
	private char Op ;
	private int funcId ;
	public static final String supported [ ] = { "sin" , "cos" , "tan" ,
		"asin" , "acos" , "atan" , "sqrt" , "abs" , "ln" , "log" , "exp" } ;

	public boolean isFunction ( ) {
		return isFunc ;
	}
	
	public boolean isVariable ( ) {
		return isX ;
	}
	
	public String toString ( ) {
		if ( isFunc ) return supported [ funcId ] ;
		if ( isX ) return "x" ;
		if ( isOp || isUnary ) return "" + Op ;
		return "" + n ;
	}

	// update the evaluation stack as appropriate
	public void eval ( Stack s , double x ) throws EmptyStackException {
		double lval , rval , ans = 0 ;
		if ( isX ) {
			s . push ( new Double ( x ) ) ;
			return ;
		} else if ( isNum ) {
			s . push ( new Double ( n ) ) ;
			return ;
		} else if ( isOp ) {
			rval = ( ( Double ) s . pop ( ) ) . doubleValue ( ) ;
			lval = ( ( Double ) s . pop ( ) ) . doubleValue ( ) ;
			if ( Op == '+' ) {
				ans = lval + rval ;
			} else if ( Op == '-' ) {
				ans = lval - rval ;
			} else if ( Op == '/' ) {
				ans = lval / rval ;
			} else if ( Op == '^' ) {
				ans = Math . pow ( lval , rval ) ;
			} else if ( Op == '*' ) {
				ans = lval * rval ;
			}
			s . push ( new Double ( ans ) ) ;
			return ;
		}
		// not a variable, constant, or operation, so must be a function
		lval = ( ( Double ) s . pop ( ) ) . doubleValue ( ) ;
		if ( isUnary ) {
			if ( Op == '+' ) {
				ans = lval ;
			} else if ( Op == '-' ) {
				ans = - lval ;
			}
		} else {
			switch ( funcId ) {
			case 0 :
				ans = Math . sin ( lval ) ;
				break ;
			case 1 :
				ans = Math . cos ( lval ) ;
				break ;
			case 2 :
				ans = Math . tan ( lval ) ;
				break ;
			case 3 :
				ans = Math . asin ( lval ) ;
				break ;
			case 4 :
				ans = Math . acos ( lval ) ;
				break ;
			case 5 :
				ans = Math . atan ( lval ) ;
				break ;
			case 6 :
				ans = Math . sqrt ( lval ) ;
				break ;
			case 7 :
				ans = Math . abs ( lval ) ;
				break ;
			case 8 :
				ans = Math . log ( lval ) ;
				break ;
			case 9 :
				ans = Math . log10 ( lval ) ;
				break ;
			case 10 :
				ans = Math . exp ( lval ) ;
				break ;
			}
		}
		s . push ( new Double ( ans ) ) ;
	}

	public MathObject ( double d ) {
		n = d ;
		isNum = true ;
	}

	public static final boolean isOperand ( char c ) {
		return ( c == '+' || c == '-' || c == '*' || c == '/' || c == '^' ) ;
	}
	
	public MathObject ( char c ) throws ParseException {
		isOp = true ;
		if ( !  isOperand ( c ) )
			throw new ParseException ( "Unrecognized token." ) ;
		Op = c ;
	}

	public MathObject ( char c , boolean unary ) throws ParseException {
		if ( ! ( c == '+' || c == '-' ) )
			throw new ParseException ( "Unrecognized token." ) ;
		Op = c ;
		if ( unary ) {
			isUnary = true ;
		} else {
			isOp = true ;
		}
	}

	public MathObject ( String s ) throws ParseException {
		int i ;
		if ( s . equals ( "x" ) ) {
			isX = true ;
			return ;
		}
		isFunc = true ;
		for ( i = 0 ; i < supported . length ; i ++ ) {
			if ( supported [ i ] . equals ( s ) ) {
				funcId = i ;
				return ;
			}
		}
		throw new ParseException ( "Unrecognized token." ) ;
	}

}
