package calculus ;
import java.util.Vector ;
import java.util.EmptyStackException ;
import java.util.Stack ;

public class TFunction {
	private String f ;
	private Vector parsed = new Vector ( 10 , 10 ) ;

	private static final int OP_NONE = 4 ;
	private static final int OP_EXP = 3 ;
	private static final int OP_TIMES = 2 ;
	private static final int OP_ADD = 1 ;

	// the constructor
	public TFunction ( String fn ) throws ParseException {
		f = fn ;
		parse ( f ) ;
	}
	
	public static final double eval ( String s ) throws ParseException {
		TFunction temp = new TFunction ( s ) ;
		for ( int i = 0 ; i < temp . parsed . size ( ) ; i ++ ) {
			if ( ( ( MathObject ) temp . parsed . get ( i ) ) . isVariable ( ) ) {
				throw new ParseException ( "Nonconstant expression." ) ;
			}
		}
		return temp . val ( 0 ) ;
	}

	public double val ( double x ) {
		Stack s = new Stack ( ) ;
		MathObject m ;
		int i ;
		try {
		// loop through the parsed list of terms, evaluate them in order
		for ( i = 0 ; i < parsed . size ( ) ; i ++ ) {
			m = ( MathObject ) parsed . get ( i ) ;
			m . eval ( s , x ) ;
		}
		if ( s . size ( ) != 1 ) {
			// this should never happen
			System . err . println ( "Evaluation error." ) ;
			System . exit ( -1 ) ;
			return 0 ;
		}
		return ( ( Double ) s .pop ( )  ) . doubleValue ( ) ;
		} catch ( EmptyStackException err ) {
			// better to return 0 than crash if malformed function
			// was not detected at parse time
			return 0 ;
		}
	}

	public double derivative ( double x ) {
		return derivative ( x , .00001 ) ;
	}

	public double derivative ( double x , double h ) {
		return ( val ( x + h ) - val ( x - h ) ) / ( 2 * h ) ;
	}

	public String getFunction ( ) {
		return f ;
	}

	public void parse ( String expression ) throws ParseException {
		String e ;
		char temp ;
		int i , eLen , splitAt = -1 ;
		boolean isUnary = true ;
		int lowestPrecedence = OP_NONE ;

		e = ( expression . trim ( ) ) . toLowerCase ( ) ;
		eLen = e . length ( ) ;
		if ( eLen == 0 )
			throw new ParseException ( "Syntax error." ) ;
		
		if ( MathObject . isOperand ( temp = e . charAt ( eLen - 1 ) ) )
			throw new ParseException ( "Binary operation " + temp + " with no right hand side." ) ;

		for ( i = 0 ; i < eLen ; i ++ ) {
			temp = e . charAt ( i ) ;
			// next character is unary operator only if this
			// is an operator, so most likely not
			if ( temp == '(' ) {
				// advance to the matching )
				int numParens = 0 ;
				boolean notDone = true ;
				for ( ; i < eLen && notDone ; i ++ ) {
					temp = e . charAt ( i ) ;
					if ( temp == '(' ) {
						numParens ++ ;
					} else if ( temp == ')' ) {
						numParens -- ;
						if ( numParens == 0 ) {
							notDone = false ;
							// have to decrease i because both loops will
							// increase it, only want it increased once
							i -- ;
						}
					}
				}
				if ( notDone ) {
					throw new ParseException ( "Imbalanced parentheses." ) ;
				}
			} else if ( temp == ')' ) {
				// should never be here except in malformed function
				throw new ParseException ( "Imbalanced parentheses." ) ;
			} else {
				// we only care about stuff not in parentheses
				if ( MathObject . isOperand ( temp ) ) {
					switch ( temp ) {
					case '+' :
					case '-' :
						if ( OP_ADD <= lowestPrecedence && ( ! isUnary ) ) {
							lowestPrecedence = OP_ADD ;
							splitAt = i ;
						}
						break ;
					case '*' :
					case '/' :
						if ( OP_TIMES <= lowestPrecedence ) {
							lowestPrecedence = OP_TIMES ;
							splitAt = i ;
							if ( isUnary )
								throw new ParseException ( "Unary " + temp + "." ) ;
						}
						break ;
					case '^' :
						if ( OP_EXP <= lowestPrecedence ) {
							lowestPrecedence = OP_EXP ;
							splitAt = i ;
							if ( isUnary )
								throw new ParseException ( "Unary " + temp + "." ) ;
						}
						break ;
					}
				}
			}
			isUnary = MathObject . isOperand ( temp ) ;
		}
		if ( lowestPrecedence != OP_NONE ) {
			// operation
			temp = e . charAt ( splitAt ) ;
			if ( splitAt == eLen - 1 )
				throw new ParseException ( "Binary operation " + temp + " with no right hand side." ) ;
			parse ( e . substring ( 0 , splitAt ) ) ;
			parse ( e . substring ( splitAt + 1 ) ) ;
			parsed . add ( new MathObject ( temp ) ) ;
		} else {
			// no (nonunary) operators
			if ( e . charAt ( 0 ) == '(' && e . charAt ( eLen - 1 ) == ')' ) {
				// ignore excess parentheses... since no operations
				// in between, the outer ones must match
				parse ( e . substring ( 1 , eLen - 1 ) ) ;
			} else if ( e . charAt ( 0 ) == '-' || e . charAt ( 0 ) == '+' ) {
				// unary negation
				if ( eLen < 2 ) {
					throw new ParseException ( "Unary operation with nothing to act on." ) ;
				}
				parse ( e . substring ( 1 ) ) ;
				parsed . add ( new MathObject ( e . charAt ( 0 ) , true ) ) ;
				return ;
			} else {
				// check about functions
				i = e . indexOf ( '(' ) ;
				if ( i > 0 ) {
					parse ( e . substring ( i ) ) ;
					parsed . add ( new MathObject ( e . substring ( 0 , i ) ) ) ;
					return ;
				} else if ( e . equals ( "x" ) ) {
					parsed . add ( new MathObject ( "x" ) ) ;
					return ;
				} else if ( e . equals ( "pi" ) ) {
					parsed . add ( new MathObject ( Math . PI ) ) ;
					return ;
				} else if ( e . equals ( "e" ) ) {
					parsed . add ( new MathObject ( Math . E ) ) ;
					return ;
				}
				// if we reach here, should be a constant.
				try {
					parsed . add ( new MathObject ( ( new Double ( e ) ) . doubleValue ( ) ) ) ;
					return ;
				} catch ( NumberFormatException err ) {
					throw new ParseException ( "Syntax error." ) ;
				}
			}
		}
	}

}
