package calculus ;
import java.awt.*;
import java.awt.event.*;

import javax.swing.*;

public class TLine extends JFrame implements ActionListener , ItemListener , MouseMotionListener , MouseListener {
	private static final long serialVersionUID = 1L;
	private static int numOpen = 0 ;
	private int width , height , oldWidth = 0 , oldHeight = 0 ;
	private int tlY1 = 0 , tlY2 = 0 , oldTLy1 = 0 , oldTLy2 = 0 ;
	private double [ ] xplot , yplot ; 
	private TFunction f ;
	private static final int NUMPTS = 2048 ; 
	private double xmin ;
	private double xmax ;
	private double ymin ;
	private double ymax ;
	private Menu fileMenu , editMenu , helpMenu ;
	private boolean hasUpdated = false ;
	private boolean clearBeforePaint = true ;
	private boolean axes ;
	private boolean isInit ;
	private boolean hideTangent = false ;
	private int numUpdates = 0 ;
	Timer t ;

	public TLine ( String function , double xMin , double yMin , double xMax , double yMax , boolean a ) {
		xplot = new double [ NUMPTS ] ;
		yplot = new double [ NUMPTS ] ;
		try {
			f = new TFunction ( function ) ;
		} catch ( ParseException err ) {
			try {
				f = new TFunction ( "0" ) ;
			} catch ( ParseException err2 ) {
				System . exit ( -1 ) ;
			}
			
		}
		axes = a ;
		xmin = xMin ;
		ymin = yMin ;
		xmax = xMax ;
		ymax = yMax ;
		isInit = false ;
		
		// force a refresh every 500 ms
		t = new Timer ( 500 , this ) ;
		t . start ( ) ;
	}
	
	public void dispose ( ) {
		super . dispose ( ) ;
		t . removeActionListener ( this ) ;
		t . stop ( ) ;
	}
	
	private double convertX ( double x ) {
		return x / width * ( xmax - xmin ) + xmin ;
	}

	private double realY ( double y ) {
		return ( 1 - ( y - ymin ) / ( ymax - ymin ) ) * height + 60 ;
	}
	
	private double realX ( double x ) {
		return ( x - xmin ) / ( xmax - xmin ) * width ;
	}

	private void init ( ) {
		width = getSize ( ) . width ;
		height = getSize ( ) . height - 70 ;

		addMouseListener ( this ) ;
		addMouseMotionListener ( this ) ;
		setBackground ( Color . white ) ;

		makeFileMenu ( ) ;
		makeEditMenu ( ) ;
		makeHelpMenu ( ) ;

		MenuBar menuBar = new MenuBar ( ) ;

		setMenuBar ( menuBar ) ;
		menuBar . add ( fileMenu ) ;
		menuBar . add ( editMenu ) ;
		menuBar . add ( helpMenu ) ;
		
		calculate ( ) ;
		isInit = true ;
	}

	public void mouseDragged ( MouseEvent e ) {
		mouseClicked ( e ) ;
	}
	
	public void mouseClicked ( MouseEvent e ) {
		int x = e . getX ( ) ;
		double x0 = convertX ( x ) ;
		double y0 = f . val ( x0 ) ;
		double m = f . derivative ( x0 , ( xmax - xmin ) * .00001 ) ;
		if ( hasUpdated ) {
			oldTLy1 = tlY1 ;
			oldTLy2 = tlY2 ;
			hasUpdated = false ;
		}
		tlY1 = ( int ) Math . round ( realY ( y0 + m * ( xmin - x0 ) ) + .5 ) ;
		tlY2 = ( int ) Math . round ( realY ( y0 + m * ( xmax - x0 ) ) + .5 ) ;
		update ( getGraphics ( ) ) ;
		e . consume ( ) ;
	}

	public void mouseReleased ( MouseEvent e ) {
	}
	
	public void mouseEntered ( MouseEvent e ) {
	}

	public void mouseExited ( MouseEvent e ) {
	}

	private void calculate ( ) {
		int i ;
		double dx ;
		dx = ( xmax - xmin ) / NUMPTS ;
		for ( i = 0 ; i < NUMPTS ; i ++ ) {
			xplot [ i ] = xmin + i * dx ;
			yplot [ i ] = f . val ( xplot [ i ] ) ;
		}
		hideTangent = true ;
	}
	
	public void update ( Graphics g ) {
		int i , x , y , lastX , lastY ;
		if ( ( ! isInit ) || g == null ) return ;
		
		// clear the window if have only drawn it a couple times
		// causes flicker, but eliminates weird gray screen bug
		numUpdates ++ ;
		if ( numUpdates < 3 ) clearBeforePaint = true ;
		
		width = getSize ( ) . width ;
		height = getSize ( ) . height - 70 ;
		g . setPaintMode ( ) ;
		g . setColor ( Color . white ) ;
		if ( width != oldWidth || height != oldHeight ) {
			hideTangent = true ;
			numUpdates = 0 ;
		}
		if ( hideTangent ) {
			tlY1 = tlY2 = 0 ;
			clearBeforePaint = true ;
			hideTangent = false ;
		}
		if ( clearBeforePaint ) {
			g . fillRect ( 0 , 0 , width , height + 70 ) ;
		}
		clearBeforePaint = false ;
		oldWidth = width ;
		oldHeight = height ;
		g . drawLine ( 0 , oldTLy1 , width , oldTLy2 ) ;
		if ( axes ) {
			g . setColor ( Color . blue ) ;
			g . drawLine ( 0 , ( int ) Math . round ( realY ( 0 ) ) , width , ( int ) Math . round ( realY ( 0 ) ) ) ;
			g . drawLine ( ( int ) Math . round ( realX ( 0 ) ) , 0 , ( int ) Math . round ( realX ( 0 ) ) , getHeight ( ) ) ;
		}
		g . setColor ( Color . black ) ;
		lastX = ( int ) Math . round ( realX ( xplot [ 0 ] ) ) ;
		lastY = ( int ) Math . round ( realY ( yplot [ 0 ] ) ) ;
		for ( i = 1 ; i < NUMPTS ; i ++ ) {
			x = ( int ) Math . round ( realX ( xplot [ i ] ) ) ;
			y = ( int ) Math . round ( realY ( yplot [ i ] ) ) ;
			g . drawLine ( lastX , lastY , x , y ) ;
			lastX = x ;
			lastY = y ;
		}
		g . setColor ( Color . red ) ;
		g . drawLine ( 0 , tlY1 , width , tlY2 ) ;
		hasUpdated = true ;
	}

	public void paint ( Graphics g ) {
		update ( getGraphics ( ) ) ;
	}

	private void makeFileMenu ( ) {
		MenuItem item;

		fileMenu = new Menu ( "File" ) ;

		item = new MenuItem ( "New" ,  new MenuShortcut( KeyEvent . VK_N ) ) ;     
		item . addActionListener ( this ) ;
		fileMenu . add ( item ) ;
	
		item = new MenuItem ( "Close" ) ;     
		item . addActionListener ( this ) ;
		fileMenu . add ( item ) ;

		fileMenu . addSeparator ( ) ;

		item = new MenuItem ( "Exit" , new MenuShortcut ( KeyEvent . VK_Q ) ) ;     
		item . addActionListener ( this ) ;
		fileMenu . add ( item ) ;
	}
	
	private void makeHelpMenu ( ) {
		MenuItem item;

		helpMenu = new Menu ( "Help" ) ;

		item = new MenuItem ( "Supported Functions" ) ;     
		item . addActionListener ( this ) ;
		helpMenu . add ( item ) ;
		
		item = new MenuItem ( "Notes" ) ;
		item . addActionListener ( this ) ;
		helpMenu . add ( item ) ;
		
		helpMenu . addSeparator ( ) ;
		
		item = new MenuItem ( "About" ) ;     
		item . addActionListener ( this ) ;
		helpMenu . add ( item ) ;
	}

	private void makeEditMenu ( ) {
		MenuItem item ;
		CheckboxMenuItem showAxes ;

		editMenu = new Menu ( "Edit" ) ;

		item = new MenuItem ( "Change Function" ) ;        
		item . addActionListener ( this ) ;
		editMenu . add ( item ) ;
		
		item = new MenuItem ( "Change View" ) ;        
		item . addActionListener ( this ) ;
		editMenu . add ( item ) ;
		
		item = new MenuItem ( "Autoscale Y" ) ;
		item . addActionListener ( this ) ;
		editMenu . add ( item ) ;
		
		showAxes = new CheckboxMenuItem ( "Show Axes" ) ;
		showAxes . setState ( axes ) ;
		showAxes . addItemListener ( this ) ;
		editMenu . add ( showAxes ) ;
	}
	
	public void itemStateChanged ( ItemEvent e ) {
		// takes for granted that there is only one; the axes selector
		axes = ! axes ;
		clearBeforePaint = true ;
		update ( getGraphics ( ) ) ;
	}

	public void actionPerformed ( ActionEvent e ) {
		String strMenuName ;
		strMenuName = e . getActionCommand ( ) ;
		
		if ( strMenuName == null ) {
			// right now, the only non-menu event is the timer
			// which should force a refresh
			update ( getGraphics ( ) ) ;
			return ;
		}

		if ( strMenuName . equals ( "Exit" ) ) {
			System . exit ( 0 ) ;
		} else if ( strMenuName . equals ( "New" ) ) {
			makeNewWindow ( f . getFunction ( ) , xmin , ymin , xmax , ymax ,
					getSize ( ) . width , getSize ( ) . height , axes ) ;
		} else if ( strMenuName . equals ( "Close" ) ) {
			dispose ( )  ;
			numOpen -- ;
			if ( numOpen <= 0 )
				System . exit ( 0 ) ;
		} else if ( strMenuName . equals ( "Change Function" ) ) {
			Object result = JOptionPane . showInputDialog ( null ,
					"Enter function to use:" , "Change Function" , JOptionPane . PLAIN_MESSAGE ,
					null , null , f . getFunction ( ) ) ;
			if ( ( String ) result != null ) {
				try {
					f = new TFunction ( ( String ) result ) ;
					calculate ( ) ;
					update ( getGraphics ( ) ) ;
				} catch ( ParseException err ) {
					JOptionPane . showMessageDialog ( null , err . getMsg ( ) + 
							"\nYour previous function is unchanged." , "Parse Error" , JOptionPane . ERROR_MESSAGE ) ;
				}
			}
		} else if ( strMenuName . equals ( "Autoscale Y" ) ) {
			ymin = ymax = yplot [ 0 ] ;
			double y ;
			for ( int i = 0 ; i < NUMPTS ; i ++ ) {
				y = yplot [ i ] ;
				if ( ! ( Double . isInfinite( y ) || Double . isNaN ( y ) ) ) {
					if ( yplot [ i ] < ymin || Double . isNaN ( ymin ) || Double . isInfinite ( ymin ) )
						ymin = yplot [ i ] ;
					if ( yplot [ i ] > ymax || Double . isNaN ( ymax ) || Double . isInfinite ( ymax ) )
						ymax = yplot [ i ] ;
				}
			}
			if ( ymin == ymax ) {
				ymin -= .00001 ;
				ymax += .00001 ;
			}
			hideTangent = true ;
			update ( getGraphics ( ) ) ;			
		} else if ( strMenuName . equals ( "Supported Functions" ) ) {
			String supported = "" ;
			for ( int i = 0 ; i < MathObject . supported . length - 1 ; i ++ ) {
				supported += MathObject . supported [ i ] + ", " ;
			}
			supported += MathObject . supported [ MathObject . supported . length - 1 ] ;
			JOptionPane . showMessageDialog ( null , "The five basic arithmetic operations are supported:  + , - , * , / , ^\n"
					+ "the constants e and pi are recognized\n"
					+ "and the following functions are supported:\n"
					+ supported , "Supported Functions" , JOptionPane . INFORMATION_MESSAGE ) ;
		} else if ( strMenuName . equals ( "About" ) ) {
			JOptionPane . showMessageDialog ( null , "TLine 1.5b\n\n"
					+ "Robert McDougal\n"
					+ "2000, 2006\n\n"
					+ "This program is provided AS IS, with NO warranties expressed or implied." ,
					"About" , JOptionPane . INFORMATION_MESSAGE ) ;
		} else if ( strMenuName . equals ( "Notes" ) ) {
			JOptionPane . showMessageDialog ( null , 
					"Expressions are valid in text fields;\n"
					+ " i.e. 1+sin(3) is a perfectly valid value for ymax." ,
					"Notes" , JOptionPane . INFORMATION_MESSAGE ) ;
		} else if ( strMenuName . equals ( "Change View" ) ) {
			JPanel p = new JPanel ( new GridLayout ( 0 , 1 ) ) ;
			JPanel p1 = new JPanel ( ) , p2 = new JPanel ( ) , p3 = new JPanel ( ) , p4 = new JPanel ( ) ;			
			JTextField xminText = new JTextField ( Double . toString ( xmin ) , 13 ) ;
			JTextField xmaxText = new JTextField ( Double . toString ( xmax ) , 13 ) ;
			JTextField yminText = new JTextField ( Double . toString ( ymin ) , 13 ) ;
			JTextField ymaxText = new JTextField ( Double . toString ( ymax ) , 13 ) ;
			JLabel xminLabel = new JLabel ( "xmin:" ) ;
			JLabel xmaxLabel = new JLabel ( "xmax:" ) ;
			JLabel yminLabel = new JLabel ( "ymin:" ) ;
			JLabel ymaxLabel = new JLabel ( "ymax:" ) ;
			p1 . add ( xminLabel ) ;
			p1 . add ( xminText ) ;
			p2 . add ( xmaxLabel ) ;
			p2 . add ( xmaxText ) ;
			p3 . add ( yminLabel ) ;
			p3 . add ( yminText ) ;
			p4 . add ( ymaxLabel ) ;
			p4 . add ( ymaxText ) ;
			p . add ( p1 ) ;
			p . add ( p2 ) ;
			p . add ( p3 ) ;
			p . add ( p4 ) ;
			int choice = JOptionPane . showConfirmDialog ( null , p , "Change View" , JOptionPane . OK_CANCEL_OPTION , JOptionPane . PLAIN_MESSAGE ) ;
			if ( choice == JOptionPane . OK_OPTION ) {
				try {
					double xminTemp = TFunction . eval ( xminText . getText ( ) ) ;
					double yminTemp = TFunction . eval ( yminText . getText ( ) ) ;
					double xmaxTemp = TFunction . eval ( xmaxText . getText ( ) ) ;
					double ymaxTemp = TFunction . eval ( ymaxText . getText ( ) ) ;
					if ( xminTemp < xmaxTemp && yminTemp < ymaxTemp ) {
						xmin = xminTemp ;
						ymin = yminTemp ;
						xmax = xmaxTemp ;
						ymax = ymaxTemp ;
						calculate ( ) ;
					} else {
						JOptionPane . showMessageDialog ( null , "Ill-defined window values.\n"
								+ "Need xmin < xmax and ymin < ymax.\n" 
								+ "Your previous window is unchanged." , "Window Error" , JOptionPane . ERROR_MESSAGE ) ;
					}
				} catch ( ParseException err ) {
					JOptionPane . showMessageDialog ( null , "Ill-defined window values.\n"
							+ err . getMsg ( ) + "\n" 
							+ "Your previous window is unchanged." , "Window Error" , JOptionPane . ERROR_MESSAGE ) ;
				}
			}
		}
	}

	public static void main ( String args [] ) {
		makeNewWindow ( "x*(x-1)*(x+1)" , -2 , -2 , 2 , 2 , 640 , 480 , true ) ;
	} ;
	
	private static void makeNewWindow ( String function , double xMin , double yMin , double xMax , double yMax , int width , int height , boolean a ) { 
		TLine window = new TLine ( function , xMin , yMin , xMax , yMax , a ) ;

		window . addWindowListener (
				new WindowAdapter( ) {
					public void windowClosing ( WindowEvent e ) {
						e . getWindow ( ) . dispose ( ) ;
						numOpen -- ;
						if ( numOpen <= 0 )
							System . exit ( 0 ) ;
					}
				}
		) ;

		window . setTitle ( "TLine" ) ;
		window . setSize ( width , height ) ;
		window . init ( ) ;
		window . setVisible ( true ) ;
		numOpen ++ ;
	}

	public void mousePressed(MouseEvent e) {
	}

	public void mouseMoved(MouseEvent e) {
	} 
}
